"""Implementation of common and variable 2-qubit controlled gates."""


import numpy as np

from gateraid.gate_base import TwoQubitGate

control_matrix = np.array([[0, 0], [0, 1]], dtype=complex)
anti_control_matrix = np.array([[1, 0], [0, 0]], dtype=complex)
I_2 = np.eye(2, dtype=complex)


class ControlledUnitary(TwoQubitGate):
    """A controlled unitary gate.

    Parameters
    ----------
    single_qubit_unitary : np.ndarray
        The single-qubit unitary to apply.
    single_qubit_unitary_name : str
        The name of the single-qubit unitary.
    first_qubit_controlled : bool, optional
        Whether the first qubit is the control qubit. Default is True.
    anti_control : bool, optional
        Whether the control is inverted. Default is False.

    Attributes
    ----------
    matrix : np.ndarray
        The unitary matrix of the gate.
    name : str
        The name of the gate.

    Examples
    -------
    >>> import numpy as np
    >>> from gateraid.controlled_gates import ControlledUnitary
    >>> X = np.array([[0, 1], [1, 0]], dtype=complex)
    >>> gate = ControlledUnitary(X, 'X')
    >>> print(gate)
    CX1->2 gate with matrix:
    [[1+0j 0+0j 0+0j 0+0j]
     [0+0j 1+0j 0+0j 0+0j]
     [0+0j 0+0j 0+0j 1+0j]
     [0+0j 0+0j 1+0j 0+0j]]

    >>> import numpy as np
    >>> from gateraid.controlled_gates import ControlledUnitary
    >>> X = np.array([[0, 1], [1, 0]], dtype=complex)
    >>> gate = ControlledUnitary(X, 'X',
                                 first_qubit_controlled=False,
                                 anti_control=True)
    >>> print(gate)
    anti-CX2->1 gate with matrix:
    [[0+0j 0+0j 1+0j 0+0j]
     [0+0j 1+0j 0+0j 0+0j]
     [1+0j 0+0j 0+0j 0+0j]
     [0+0j 0+0j 0+0j 1+0j]]
    """

    def __init__(self,  # noqa D107
                 single_qubit_unitary: np.ndarray,
                 single_qubit_unitary_name: str,
                 first_qubit_controlled: bool = True,
                 anti_control: bool = False):
        control_a, control_b = control_matrix, anti_control_matrix
        if anti_control:
            # Swap the control matrices
            control_a, control_b = control_b, control_a
        two_qubit_unitary = np.kron(control_a, single_qubit_unitary) + \
            np.kron(control_b, I_2) if first_qubit_controlled else \
            np.kron(single_qubit_unitary, control_a) + np.kron(I_2, control_b)

        name = f'{"anti-" if anti_control else ""}' \
            f'C{single_qubit_unitary_name}'\
            f'{"1->2" if first_qubit_controlled else "2->1"}'

        super().__init__(matrix=two_qubit_unitary, name=name)


def make_CX(first_qubit_controlled=True, anti_control=False):
    """Define the controlled-X gate.

    Example
    -------
    >>> import numpy as np
    >>> from gateraid.controlled_gates import make_CX
    >>> from gateraid.quantum_state import QuantumState
    >>> state = QuantumState(np.array([0, 0, 1, 0]))
    >>> CX = make_CX()
    >>> new_state = CX(state)
    >>> print(new_state)
    State:
    0+0j|00> +
    0+0j|01> +
    0+0j|10> +
    1+0j|11>
    """
    return ControlledUnitary(single_qubit_unitary=np.matrix([[0, 1], [1, 0]],
                                                            dtype=complex),
                             single_qubit_unitary_name='X',
                             first_qubit_controlled=first_qubit_controlled,
                             anti_control=anti_control)


def make_CY(first_qubit_controlled=True, anti_control=False):
    """Define the controlled-Y gate.

    Example
    -------
    >>> import numpy as np
    >>> from gateraid.controlled_gates import make_CY
    >>> from gateraid.quantum_state import QuantumState
    >>> state = QuantumState(np.array([0, 0, 1, 0]))
    >>> CY = make_CY()
    >>> new_state = CY(state)
    >>> print(new_state)
    State:
    0+0j|00> +
    0+0j|01> +
    0+0j|10> +
    0+1j|11>
    """
    return ControlledUnitary(single_qubit_unitary=np.matrix([[0, -1j],
                                                             [1j, 0]],
                                                            dtype=complex),
                             single_qubit_unitary_name='Y',
                             first_qubit_controlled=first_qubit_controlled,
                             anti_control=anti_control)


def make_CZ(first_qubit_controlled=True, anti_control=False):
    """Define the controlled-Z gate.

    Example
    -------
    >>> import numpy as np
    >>> from gateraid.controlled_gates import make_CZ
    >>> from gateraid.quantum_state import QuantumState
    >>> state = QuantumState(np.array([0, 0, 0, 1]))
    >>> CZ = make_CZ()
    >>> new_state = CZ(state)
    >>> print(new_state)
    State:
    0+0j|00> +
    0+0j|01> +
    0+0j|10> +
    -1+0j|11>
    """
    return ControlledUnitary(single_qubit_unitary=np.matrix([[1, 0], [0, -1]],
                                                            dtype=complex),
                             single_qubit_unitary_name='Z',
                             first_qubit_controlled=first_qubit_controlled,
                             anti_control=anti_control)



"""Utilities for gateraid."""


import numpy as np


def check_unitary(matrix: np.ndarray) -> None:
    """Verify that a matrix is unitary.

    Parameters
    ----------
    matrix : np.ndarray
        The matrix to verify.

    Raises
    ------
    TypeError
        If the matrix is not a numpy array.
    TypeError
        If the matrix is not of dtype float, complex or int.
    ValueError
        If the matrix shape is not (4, 4).
    ValueError
        If the matrix is not unitary.
    """
    if not isinstance(matrix, np.ndarray):
        raise TypeError("Matrix should be a 4x4 numpy array, "
                        f"but a {type(matrix).__name__} was provided.")
    if matrix.shape != (4, 4):
        raise ValueError(f"Matrix has shape {matrix.shape}, "
                         "but should be (4,4).")
    if matrix.dtype not in (complex, float, int):
        raise TypeError(f"Matrix has dtype {matrix.dtype}, "
                        f"but should be float, complex or int.")
    if not np.allclose(matrix.conj().T @ matrix, np.eye(4)):
        raise ValueError("Input matrix is not unitary. Should have U†U = I, "
                         f"but U†U =\n{matrix.conj().T @ matrix}")



"""Implementation of a 2-qubit quantum state."""


import numpy as np

from gateraid.utilities import check_unitary


class QuantumState:
    """A 2-qubit quantum state.

    Parameters
    ----------
    state : np.ndarray, optional
        The state to initialize. Default is |00>=np.array([1, 0, 0, 0]).

    Attributes
    ----------
    state : np.ndarray
        The state of the quantum system.

    Raises
    ------
    ValueError
        If the state is not a 4x1 column vector.
    ValueError
        If the state is not normalized.

    Methods
    -------
    __str__()
        Return string representation of the quantum state.
    __repr__()
        Return canonical string representation of the quantum state.
    apply_matrix(matrix: np.ndarray, _skip_validation=False) -> QuantumState
        Apply a matrix to the state.

    Examples
    --------
    >>> from gateraid.quantum_state import QuantumState
    >>> state = QuantumState()
    >>> print(state)
    State:
    1+0j|00> +
    0+0j|01> +
    0+0j|10> +
    0+0j|11>

    >>> import numpy as np
    >>> from gateraid.quantum_state import QuantumState
    >>> state = QuantumState(np.array([1, 0, 0, 1])/np.sqrt(2))
    >>> print(state)
    State:
    0.7071+0j|00> +
    0+0j|01> +
    0+0j|10> +
    0.7071+0j|11>
    """

    def __init__(self, state=None):  # noqa: D107
        if state is None:
            state = np.array([1, 0, 0, 0], dtype=complex)
        state = state.astype(complex)
        if state.shape == (4,):
            # Make the state a column vector suitable for matrix multiplication
            state = state[:, None]
        if state.shape != (4, 1):
            raise ValueError(f"State has shape {state.shape}, "
                             "but it should be (4,1) or (4,)")
        if not np.allclose(np.linalg.norm(state), 1):
            raise ValueError("Provided state has a norm of "
                             f"{np.linalg.norm(state)}, but "
                             "this should be (approximately) 1.")
        self.state = state

    def __repr__(self):
        """Return canonical string representation of the quantum state."""
        return f"{self.__class__.__name__}({self.state!r})"

    def __str__(self):
        """Return string representation of the quantum state."""
        return f"State: \
            \n{self.state[0, 0]:.4g}|00> + \n{self.state[1, 0]:.4g}|01> + \
            \n{self.state[2, 0]:.4g}|10> + \n{self.state[3, 0]:.4g}|11>"

    def apply_matrix(self, matrix: np.ndarray, _skip_validation=False):
        """Apply a matrix to the state.

        Parameters
        ----------
        matrix : np.ndarray
            The matrix to apply.
        _skip_validation : bool, optional
            Skip the validation of the matrix, by default False.

        Returns
        -------
        QuantumState
            The updated quantum state.

        Example
        -------
        >>> state = QuantumState(np.array([0, 0, 1, 0]))
        >>> matrix = np.array([[1, 0, 0, 0],
                               [0, 1, 0, 0],
                               [0, 0, 0, 1],
                               [0, 0, 1, 0]], dtype=complex)
        >>> state.apply_matrix(matrix)
        >>> print(state)
        State:
        0+0j|00> +
        0+0j|01> +
        0+0j|10> +
        1+0j|11>
        """
        if not _skip_validation:
            # If applying this using the Gate(State) notation,
            # no need to repeat the checks
            check_unitary(matrix)
        self.state = matrix @ self.state
        return self



"""Implementation of a general 2-qubit unitary gate."""


from abc import ABC
import numpy as np

from gateraid.utilities import check_unitary
from gateraid.quantum_state import QuantumState

np.set_printoptions(formatter={'all': lambda x: "{:.4g}".format(x)})


class TwoQubitGate(ABC):
    """A 2-qubit unitary gate.

    Parameters
    ----------
    matrix : np.ndarray
        The unitary matrix of the gate.
    name : str
        The name of the gate.

    Attributes
    ----------
    matrix : np.ndarray
        The unitary matrix of the gate.
    name : str
        The name of the gate.

    Methods
    -------
    __call__(state: QuantumState) -> QuantumState
        Apply the gate to a quantum state.
    __str__()
        Return string representation of the gate.
    __repr__()
        Return canonical string representation of the gate.

    Example
    -------
    >>> import numpy as np
    >>> from gateraid.gate_base import TwoQubitGate
    >>> matrix = np.array([[1, 0, 0, 0],
                           [0, 1, 0, 0],
                           [0, 0, 0, 1],
                           [0, 0, 1, 0]], dtype=complex)
    >>> name = 'CNOT'
    >>> gate = TwoQubitGate(matrix, name)
    >>> print(gate)
    CNOT gate with matrix:
    [[1+0j 0+0j 0+0j 0+0j]
     [0+0j 1+0j 0+0j 0+0j]
     [0+0j 0+0j 0+0j 1+0j]
     [0+0j 0+0j 1+0j 0+0j]]
    """

    def __init__(self, matrix: np.ndarray, name: str):  # noqa: D107
        self.matrix = matrix
        check_unitary(matrix)
        self.name = name

    def __call__(self, state: QuantumState) -> QuantumState:
        """Apply the gate to a quantum state."""
        return state.apply_matrix(self.matrix, _skip_validation=True)

    def __str__(self):
        """Return string representation of the gate."""
        return f"{self.name} gate with matrix: \n{self.matrix}"

    def __repr__(self):
        """Return canonical string representation of the gate."""
        return f"{self.__class__.__name__}(\
            matrix={self.matrix!r}, name={self.name!r})"



import numpy as np
from scipy.linalg import expm
from gateraid.other_gates import *
from gateraid.controlled_gates import *

def test_GeneralUnitary():
    theta = 3.5
    input_1 = {'IZ': theta}
    input_2 = {'ZZ': theta}
    Z = make_Z()
    single_qubit_unitary = expm(theta*1j*Z)
    U0 = LocalUnitary(single_qubit_unitary, f"exp(i{theta}Z)", 1)
    U1 = GeneralUnitary(input_1)
    U2 = GeneralUnitary(input_2)
    assert np.allclose(U0.matrix, U1.matrix) 
    assert not np.allclose(U0.matrix, U2.matrix) 

def test_SWAP():
    Rxx = GeneralUnitary({'XX': -np.pi/4}).matrix
    Ryy = GeneralUnitary({'YY': -np.pi/4}).matrix
    Rzz = GeneralUnitary({'ZZ': -np.pi/4}).matrix
    SWAP_1 = (1+1j)/np.sqrt(2) * Rxx @ Ryy @ Rzz
    SWAP_2 = make_SWAP().matrix
    assert np.allclose(SWAP_1, SWAP_2)

def test_iSWAP():
    Rxx = GeneralUnitary({'XX': np.pi/4}).matrix
    Ryy = GeneralUnitary({'YY': np.pi/4}).matrix
    iSWAP_1 = Rxx @ Ryy 
    iSWAP_2 = make_iSWAP().matrix
    assert np.allclose(iSWAP_1, iSWAP_2)

def test_sqrt_SWAP():
    input = {'II': np.pi/8, 'XX': -np.pi/8, 'YY': -np.pi/8, 'ZZ': -np.pi/8}
    sqrt_SWAP_1 = GeneralUnitary(input).matrix
    sqrt_SWAP_2 = make_sqrt_SWAP().matrix
    assert np.allclose(sqrt_SWAP_1, sqrt_SWAP_2)

def test_sqrt_iSWAP():
    Rxx = GeneralUnitary({'XX': np.pi/8}).matrix
    Ryy = GeneralUnitary({'YY': np.pi/8}).matrix
    sqrt_iSWAP_1 = Rxx @ Ryy 
    sqrt_iSWAP_2 = make_sqrt_iSWAP().matrix
    assert np.allclose(sqrt_iSWAP_1, sqrt_iSWAP_2)

def test_CNOT():
    Ry1 = GeneralUnitary({'YI': np.pi/4}).matrix
    Rx1 = GeneralUnitary({'XI': np.pi/4}).matrix
    Rx2 = GeneralUnitary({'IX': np.pi/4}).matrix
    Rxx = GeneralUnitary({'XX': -np.pi/4}).matrix
    Ry2 = GeneralUnitary({'YI': -np.pi/4}).matrix
    CNOT_1 = (1-1j)/np.sqrt(2) * Ry1 @ Rx1 @ Rx2 @ Rxx @ Ry2
    CNOT_2 = make_CX().matrix
    assert np.allclose(CNOT_1, CNOT_2)

def test_CZ():
    CNOT = make_CX().matrix
    H = make_H()
    IH = LocalUnitary(H, 'H', 1).matrix
    CZ_1 = IH @ CNOT @ IH
    CZ_2 = make_CZ().matrix
    assert np.allclose(CZ_1, CZ_2)



"""Implementation of common and variable 2-qubit gates."""


import numpy as np

from scipy.linalg import expm
from gateraid.gate_base import TwoQubitGate


class GeneralUnitary(TwoQubitGate):
    """Initialise a 2-qubit gate, in the form U = exp(i[∑_i c_i P_i]), where \
    P_i are Pauli matrices and c_i the coefficients.

    Parameters
    ----------
    pauli_dict : dict[str, float]
        A dictionary where the keys represent 2-qubit Pauli matricies
        and the values are the coefficients of the Pauli matrices.

    Attributes
    ----------
    matrix : np.ndarray
        The unitary matrix of the gate.
    name : str
        The name of the gate.

    Raises
    ------
    ValueError
        If the keys of the dictionary are not 2-character strings, made up of
        'I', 'X', 'Y' and 'Z'.
    ValueError
        If the values of the dictionary are not real numbers.

    Example
    -------
    >>> import numpy as np
    >>> from gateraid.other_gates import GeneralUnitary
    >>> pauli_dict = {'II': 1, 'XX': 1, 'YY': 1, 'ZZ': 1}
    >>> gate = GeneralUnitary(pauli_dict)
    >>> print(gate)
    exp(i[+1II+1XX+1YY+1ZZ]) gate with matrix:
    [[-0.4161+0.9093j 0+0j 0+0j 0+0j]
     [0+0j -0.4161+0j 0+0.9093j 0+0j]
     [0+0j 0+0.9093j -0.4161+0j 0+0j]
     [0+0j 0+0j 0+0j -0.4161+0.9093j]]
    """

    def __init__(self,  # noqa D107
                 pauli_dict: dict[str, float]):

        # Lookup dictionary for the 1-qubit Paulis
        lookup = {'I': np.eye(2, dtype=complex),
                  'X': np.array([[0, 1], [1, 0]], dtype=complex),
                  'Y': np.array([[0, -1j], [1j, 0]], dtype=complex),
                  'Z': np.array([[1, 0], [0, -1]], dtype=complex)}

        name = 'exp(i['

        # Generate a linear combination of 2-qubit Pauli strings
        operator = np.zeros((4, 4), dtype=complex)
        for key in pauli_dict:
            paulis = list(key)
            if (
                (len(paulis) != 2)
                or (paulis[0] not in lookup.keys())
                or (paulis[1] not in lookup.keys())
            ):
                raise ValueError("Keys of dictionary should be 2-character "
                                 "strings where the characters are either "
                                 "'I', 'X', 'Y' or 'Z'.")
            pauli_string = np.kron(lookup[paulis[0]], lookup[paulis[1]])

            coeff = pauli_dict[key]
            if not np.isreal(coeff):
                raise ValueError("Values of dictionary should be real "
                                 "numbers.")

            operator += coeff * pauli_string
            name += f'+{round(coeff, 2)}{key}'

        name += '])'

        # Take the exponential of this hermitian operator times
        # the imaginary unit
        two_qubit_unitary = expm(1j*operator)

        super().__init__(matrix=two_qubit_unitary, name=name)


class LocalUnitary(TwoQubitGate):
    """Initialise a 2-qubit gate that acts on a single qubit.

    Parameters
    ----------
    single_qubit_unitary : np.ndarray
        The single-qubit unitary that acts on the specified site.
    single_qubit_unitary_name : str
        The name of the single-qubit unitary.
    site : int
        The index of the qubit that the single-qubit unitary acts on.

    Attributes
    ----------
    matrix : np.ndarray
        The unitary matrix of the gate.
    name : str
        The name of the gate.

    Raises
    ------
    ValueError
        If the site index is not 0 or 1.

    Example
    -------
    >>> import numpy as np
    >>> from gateraid.other_gates import LocalUnitary
    >>> X = np.array([[0, 1], [1, 0]], dtype=complex)
    >>> gate = LocalUnitary(X, 'X', 0)
    >>> print(gate)
    X_I gate with matrix:
    [[0+0j 0+0j 1+0j 0+0j]
     [0+0j 0+0j 0+0j 1+0j]
     [1+0j 0+0j 0+0j 0+0j]
     [0+0j 1+0j 0+0j 0+0j]]
    """

    def __init__(self,  # noqa D107
                 single_qubit_unitary: np.ndarray,
                 single_qubit_unitary_name: str,
                 site: int):
        I = np.eye(2, dtype=complex)  # noqa: E741, N806
        if site not in [0, 1]:
            raise ValueError("Site index is out of range - "
                             "should take values of 0 or 1.")

        # Take tensor product of the single-qubit unitary on specifiied site
        # with the identity on unused site
        elif site == 0:
            two_qubit_unitary = np.kron(single_qubit_unitary, I)
            name = single_qubit_unitary_name + '_I'
        else:
            two_qubit_unitary = np.kron(I, single_qubit_unitary)
            name = 'I_' + single_qubit_unitary_name

        super().__init__(matrix=two_qubit_unitary, name=name)


def make_SWAP():
    """Define the SWAP gate.

    Example
    -------
    >>> import numpy as np
    >>> from gateraid.other_gates import make_SWAP
    >>> from gateraid.quantum_state import QuantumState
    >>> state = QuantumState(np.array([0, 1, 0, 0]))
    >>> SWAP = make_SWAP()
    >>> new_state = SWAP(state)
    >>> print(new_state)
    State:
    0+0j|00> +
    0+0j|01> +
    1+0j|10> +
    0+0j|11>
    """
    matrix = np.array([[1, 0, 0, 0],
                       [0, 0, 1, 0],
                       [0, 1, 0, 0],
                       [0, 0, 0, 1]], dtype=complex)
    return TwoQubitGate(matrix=matrix, name='SWAP')


def make_iSWAP():
    """Define the iSWAP gate.

    Example
    -------
    >>> import numpy as np
    >>> from gateraid.other_gates import make_iSWAP
    >>> from gateraid.quantum_state import QuantumState
    >>> state = QuantumState(np.array([0, 1, 0, 0]))
    >>> iSWAP = make_iSWAP()
    >>> new_state = iSWAP(state)
    >>> print(new_state)
    State:
    0+0j|00> +
    0+0j|01> +
    0+1j|10> +
    0+0j|11>
    """
    matrix = np.array([[1, 0, 0, 0],
                       [0, 0, 1j, 0],
                       [0, 1j, 0, 0],
                       [0, 0, 0, 1]], dtype=complex)
    return TwoQubitGate(matrix=matrix, name='iSWAP')


def make_sqrt_SWAP():
    """Define the √SWAP gate.

    Example
    -------
    >>> import numpy as np
    >>> from gateraid.other_gates import make_sqrt_SWAP
    >>> from gateraid.quantum_state import QuantumState
    >>> state = QuantumState(np.array([0, 1, 0, 0]))
    >>> sqrt_SWAP = make_sqrt_SWAP()
    >>> new_state = sqrt_SWAP(state)
    >>> print(new_state)
    State:
    0+0j|00> +
    0.5+0.5j|01> +
    0.5-0.5j|10> +
    0+0j|11>
    """
    matrix = np.array([[1, 0, 0, 0],
                       [0, (1+1j)/2, (1-1j)/2, 0],
                       [0, (1-1j)/2, (1+1j)/2, 0],
                       [0, 0, 0, 1]], dtype=complex)
    return TwoQubitGate(matrix=matrix, name='√SWAP')


def make_sqrt_iSWAP():
    """Define the √iSWAP gate.

    Example
    -------
    >>> import numpy as np
    >>> from gateraid.other_gates import make_sqrt_iSWAP
    >>> from gateraid.quantum_state import QuantumState
    >>> state = QuantumState(np.array([0, 1, 0, 0]))
    >>> sqrt_iSWAP = make_sqrt_iSWAP()
    >>> new_state = sqrt_iSWAP(state)
    >>> print(new_state)
    State:
    0+0j|00> +
    0.7071+0j|01> +
    0+0.7071j|10> +
    0+0j|11>
    """
    matrix = np.array([[1, 0, 0, 0],
                       [0, 1/np.sqrt(2), 1j/np.sqrt(2), 0],
                       [0, 1j/np.sqrt(2), 1/np.sqrt(2), 0],
                       [0, 0, 0, 1]], dtype=complex)
    return TwoQubitGate(matrix=matrix, name='√iSWAP')


def make_X():
    """Define the X gate.

    Example
    -------
    >>> from gateraid.other_gates import make_X, LocalUnitary
    >>> from gateraid.quantum_state import QuantumState
    >>> state = QuantumState()
    >>> X = make_X()
    >>> XI = LocalUnitary(X, 'X', 0)
    >>> new_state = XI(state)
    >>> print(new_state)
    State:
    0+0j|00> +
    0+0j|01> +
    1+0j|10> +
    0+0j|11>
    """
    return np.array([[0, 1], [1, 0]], dtype=complex)


def make_Y():
    """Define the Y gate.

    Example
    -------
    >>> from gateraid.other_gates import make_Y, LocalUnitary
    >>> from gateraid.quantum_state import QuantumState
    >>> state = QuantumState()
    >>> Y = make_Y()
    >>> YI = LocalUnitary(Y, 'Y', 0)
    >>> new_state = YI(state)
    >>> print(new_state)
    State:
    0+0j|00> +
    0+0j|01> +
    0+1j|10> +
    0+0j|11>
    """
    return np.array([[0, -1j], [1j, 0]], dtype=complex)


def make_Z():
    """Define the Z gate.

    Example
    -------
    >>> import numpy as np
    >>> from gateraid.other_gates import make_Z, LocalUnitary
    >>> from gateraid.quantum_state import QuantumState
    >>> state = QuantumState(np.array([0, 0, 1, 0]))
    >>> Z = make_Z()
    >>> ZI = LocalUnitary(Z, 'Z', 0)
    >>> new_state = ZI(state)
    >>> print(new_state)
    State:
    0+0j|00> +
    0+0j|01> +
    -1+0j|10> +
    0+0j|11>
    """
    return np.array([[1, 0], [0, -1]], dtype=complex)


def make_H():
    """Define the Hadamard gate.

    Example
    -------
    >>> from gateraid.other_gates import make_H, LocalUnitary
    >>> from gateraid.quantum_state import QuantumState
    >>> state = QuantumState()
    >>> H = make_H()
    >>> HI = LocalUnitary(H, 'H', 0)
    >>> new_state = HI(state)
    >>> print(new_state)
    State:
    0.7071+0j|00> +
    0+0j|01> +
    0.7071+0j|10> +
    0+0j|11>
    """
    return np.array([[1, 1], [1, -1]], dtype=complex)/np.sqrt(2)


def make_S():
    """Define the S gate.

    Example
    -------
    >>> import numpy as np
    >>> from gateraid.other_gates import make_S, LocalUnitary
    >>> from gateraid.quantum_state import QuantumState
    >>> state = QuantumState(np.array([0, 0, 1, 0]))
    >>> S = make_S()
    >>> SI = LocalUnitary(S, 'S', 0)
    >>> new_state = SI(state)
    >>> print(new_state)
    State:
    0+0j|00> +
    0+0j|01> +
    0+1j|10> +
    0+0j|11>
    """
    return np.array([[1, 0], [0, 1j]], dtype=complex)


def make_T():
    """Define the T gate.

    Example
    -------
    >>> import numpy as np
    >>> from gateraid.other_gates import make_T, LocalUnitary
    >>> from gateraid.quantum_state import QuantumState
    >>> state = QuantumState(np.array([0, 0, 1, 0]))
    >>> T = make_T()
    >>> TI = LocalUnitary(T, 'T', 0)
    >>> new_state = TI(state)
    >>> print(new_state)
    State:
    0+0j|00> +
    0+0j|01> +
    0.7071+0.7071j|10> +
    0+0j|11>
    """
    return np.array([[1, 0], [0, (1+1j)/np.sqrt(2)]], dtype=complex)



from setuptools import setup, find_packages  # noqa: D100

setup(
    name="GaterAid",
    version="1.0.2",
    packages=find_packages(),
    install_requires=[
        "numpy >= 2.0.0",
    ],
    author="Eleanor Kneip, Stasiu Wolanski, Joel the intern",
    author_email="eleanor.kneip.24@ucl.ac.uk, \
        stanislaw.wolanski.24@ucl.ac.uk, joel.mills.24@ucl.ac.uk",
    description="A package to abate the strait of applying a two-qubit gate \
        to a quantum state.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    project_urls={
        "Documentation": "https://statuesque-smakager-de86a9.netlify.app",  # or whatever URL you end up using
        "Source": "https://github.com/QC2-python-SE/QWACOBEC",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



import numpy as np  # noqa: D104

from .gate_base import TwoQubitGate  # noqa: F401
from .quantum_state import QuantumState  # noqa: F401
from .controlled_gates import *  # noqa: F401, F403
from .other_gates import *  # noqa: F401, F403

# Set numpy printing options for nicer matrix printing
np.set_printoptions(formatter={'all': lambda x: "{:.4g}".format(x)})
